@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2012  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2013  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -45,13 +45,19 @@ struct connman_config_service {
 	unsigned int ssid_len;
 	char *eap;
 	char *identity;
+	char *anonymous_identity;
 	char *ca_cert_file;
+	char *subject_match;
+	char *altsubject_match;
+	char *domain_suffix_match;
+	char *domain_match;
 	char *client_cert_file;
 	char *private_key_file;
 	char *private_key_passphrase;
 	char *private_key_passphrase_type;
 	char *phase2;
 	char *passphrase;
+	enum connman_service_security security;
 	GSList *service_identifiers;
 	char *config_ident; /* file prefix */
 	char *config_entry; /* entry name */
@@ -66,6 +72,7 @@ struct connman_config_service {
 	char *ipv6_gateway;
 	char *ipv6_privacy;
 	char *mac;
+	bool mdns;
 	char **nameservers;
 	char **search_domains;
 	char **timeservers;
@@ -97,9 +104,16 @@ static bool cleanup = false;
 #define SERVICE_KEY_PRV_KEY_PASS       "PrivateKeyPassphrase"
 #define SERVICE_KEY_PRV_KEY_PASS_TYPE  "PrivateKeyPassphraseType"
 #define SERVICE_KEY_IDENTITY           "Identity"
+#define SERVICE_KEY_ANONYMOUS_IDENTITY "AnonymousIdentity"
+#define SERVICE_KEY_SUBJECT_MATCH      "SubjectMatch"
+#define SERVICE_KEY_ALT_SUBJECT_MATCH  "AltSubjectMatch"
+#define SERVICE_KEY_DOMAIN_SUFF_MATCH  "DomainSuffixMatch"
+#define SERVICE_KEY_DOMAIN_MATCH       "DomainMatch"
 #define SERVICE_KEY_PHASE2             "Phase2"
 #define SERVICE_KEY_PASSPHRASE         "Passphrase"
+#define SERVICE_KEY_SECURITY           "Security"
 #define SERVICE_KEY_HIDDEN             "Hidden"
+#define SERVICE_KEY_MDNS               "mDNS"
 
 #define SERVICE_KEY_IPv4               "IPv4"
 #define SERVICE_KEY_IPv6               "IPv6"
@@ -127,13 +141,20 @@ static const char *service_possible_keys[] = {
 	SERVICE_KEY_PRV_KEY_PASS,
 	SERVICE_KEY_PRV_KEY_PASS_TYPE,
 	SERVICE_KEY_IDENTITY,
+	SERVICE_KEY_ANONYMOUS_IDENTITY,
+	SERVICE_KEY_SUBJECT_MATCH,
+	SERVICE_KEY_ALT_SUBJECT_MATCH,
+	SERVICE_KEY_DOMAIN_SUFF_MATCH,
+	SERVICE_KEY_DOMAIN_MATCH,
 	SERVICE_KEY_PHASE2,
 	SERVICE_KEY_PASSPHRASE,
+	SERVICE_KEY_SECURITY,
 	SERVICE_KEY_HIDDEN,
 	SERVICE_KEY_IPv4,
 	SERVICE_KEY_IPv6,
 	SERVICE_KEY_IPv6_PRIVACY,
 	SERVICE_KEY_MAC,
+	SERVICE_KEY_MDNS,
 	SERVICE_KEY_NAMESERVERS,
 	SERVICE_KEY_SEARCH_DOMAINS,
 	SERVICE_KEY_TIMESERVERS,
@@ -175,7 +196,7 @@ static void unregister_service(gpointer data)
 							list = list->next) {
 		service_id = list->data;
 
-		service = __connman_service_lookup_from_ident(service_id);
+		service = connman_service_lookup_from_identifier(service_id);
 		if (service) {
 			__connman_service_set_immutable(service, false);
 			__connman_service_set_config(service, NULL, NULL);
@@ -217,7 +238,12 @@ free_only:
 	g_free(config_service->ssid);
 	g_free(config_service->eap);
 	g_free(config_service->identity);
+	g_free(config_service->anonymous_identity);
 	g_free(config_service->ca_cert_file);
+	g_free(config_service->subject_match);
+	g_free(config_service->altsubject_match);
+	g_free(config_service->domain_suffix_match);
+	g_free(config_service->domain_match);
 	g_free(config_service->client_cert_file);
 	g_free(config_service->private_key_file);
 	g_free(config_service->private_key_passphrase);
@@ -399,7 +425,7 @@ static bool load_service_generic(GKeyFile *keyfile,
 			char *ptr;
 			long int value = strtol(mask, &ptr, 10);
 
-			if (ptr != mask && *ptr == '\0' && value <= 32)
+			if (ptr != mask && *ptr == '\0' && value && value <= 32)
 				prefix_len = value;
 
 			addr = 0xffffffff << (32 - prefix_len);
@@ -491,6 +517,9 @@ static bool load_service_generic(GKeyFile *keyfile,
 			g_strfreev(strlist);
 	}
 
+	service->mdns = __connman_config_get_bool(keyfile, group,
+						SERVICE_KEY_MDNS, NULL);
+
 	return true;
 
 err:
@@ -513,6 +542,7 @@ static bool load_service(GKeyFile *keyfile, const char *group,
 	struct connman_config_service *service;
 	const char *ident;
 	char *str, *hex_ssid;
+	enum connman_service_security security;
 	bool service_created = false;
 
 	/* Strip off "service_" prefix */
@@ -540,8 +570,8 @@ static bool load_service(GKeyFile *keyfile, const char *group,
 		g_free(service->type);
 		service->type = str;
 	} else {
-		DBG("Type of the configured service is missing for group %s",
-									group);
+		connman_warn("Type of the configured service is missing "
+			"for group %s",	group);
 		goto err;
 	}
 
@@ -651,6 +681,41 @@ static bool load_service(GKeyFile *keyfile, const char *group,
 		service->identity = str;
 	}
 
+	str = __connman_config_get_string(keyfile, group,
+					SERVICE_KEY_ANONYMOUS_IDENTITY, NULL);
+	if (str) {
+		g_free(service->anonymous_identity);
+		service->anonymous_identity = str;
+	}
+
+	str = __connman_config_get_string(keyfile, group,
+					SERVICE_KEY_SUBJECT_MATCH, NULL);
+	if (str) {
+		g_free(service->subject_match);
+		service->subject_match = str;
+	}
+
+	str = __connman_config_get_string(keyfile, group,
+					SERVICE_KEY_ALT_SUBJECT_MATCH, NULL);
+	if (str) {
+		g_free(service->altsubject_match);
+		service->altsubject_match = str;
+	}
+
+	str = __connman_config_get_string(keyfile, group,
+					SERVICE_KEY_DOMAIN_SUFF_MATCH, NULL);
+	if (str) {
+		g_free(service->domain_suffix_match);
+		service->domain_suffix_match = str;
+	}
+
+	str = __connman_config_get_string(keyfile, group,
+					SERVICE_KEY_DOMAIN_MATCH, NULL);
+	if (str) {
+		g_free(service->domain_match);
+		service->domain_match = str;
+	}
+
 	str = __connman_config_get_string(keyfile, group, SERVICE_KEY_PHASE2, NULL);
 	if (str) {
 		g_free(service->phase2);
@@ -664,6 +729,49 @@ static bool load_service(GKeyFile *keyfile, const char *group,
 		service->passphrase = str;
 	}
 
+	str = __connman_config_get_string(keyfile, group, SERVICE_KEY_SECURITY,
+			NULL);
+	security = __connman_service_string2security(str);
+
+	if (service->eap) {
+
+		if (str && security != CONNMAN_SERVICE_SECURITY_8021X)
+			connman_info("Mismatch between EAP configuration and "
+					"setting %s = %s",
+					SERVICE_KEY_SECURITY, str);
+
+		service->security = CONNMAN_SERVICE_SECURITY_8021X;
+
+	} else if (service->passphrase) {
+
+		if (str) {
+			if (security == CONNMAN_SERVICE_SECURITY_PSK ||
+					security == CONNMAN_SERVICE_SECURITY_WEP) {
+				service->security = security;
+			} else {
+				connman_info("Mismatch with passphrase and "
+						"setting %s = %s",
+						SERVICE_KEY_SECURITY, str);
+
+				service->security =
+					CONNMAN_SERVICE_SECURITY_PSK;
+			}
+
+		} else
+			service->security = CONNMAN_SERVICE_SECURITY_PSK;
+	} else if (str) {
+
+		if (security != CONNMAN_SERVICE_SECURITY_NONE)
+			connman_info("Mismatch no security and "
+					"setting %s = %s",
+					SERVICE_KEY_SECURITY, str);
+
+		service->security = CONNMAN_SERVICE_SECURITY_NONE;
+	} else
+		service->security = CONNMAN_SERVICE_SECURITY_NONE;
+
+	g_free(str);
+
 	service->config_ident = g_strdup(config->ident);
 	service->config_entry = g_strdup_printf("service_%s", service->ident);
 
@@ -700,8 +808,11 @@ static bool load_service_from_keyfile(GKeyFile *keyfile,
 	groups = g_key_file_get_groups(keyfile, NULL);
 
 	for (i = 0; groups[i]; i++) {
-		if (!g_str_has_prefix(groups[i], "service_"))
+		if (!g_str_has_prefix(groups[i], "service_")) {
+			connman_warn("Ignore group named '%s' because prefix "
+				"is not 'service_'", groups[i]);
 			continue;
+		}
 		if (load_service(keyfile, groups[i], config))
 			found = true;
 	}
@@ -855,10 +966,10 @@ static void config_notify_handler(struct inotify_event *event,
 		return;
 	}
 
-	if (event->mask & IN_CREATE)
+	if (event->mask & (IN_CREATE | IN_MOVED_TO))
 		create_config(ident);
 
-	if (event->mask & IN_MODIFY) {
+	if (event->mask & (IN_MODIFY | IN_MOVED_TO)) {
 		struct connman_config *config;
 
 		config = g_hash_table_lookup(config_table, ident);
@@ -880,7 +991,7 @@ static void config_notify_handler(struct inotify_event *event,
 		}
 	}
 
-	if (event->mask & IN_DELETE)
+	if (event->mask & (IN_DELETE | IN_MOVED_FROM))
 		g_hash_table_remove(config_table, ident);
 }
 
@@ -917,6 +1028,11 @@ char *__connman_config_get_string(GKeyFile *key_file,
 	if (!str)
 		return NULL;
 
+	/* passphrases can have spaces in the end */
+	if (!g_strcmp0(key, SERVICE_KEY_PASSPHRASE) ||
+			!g_strcmp0(key, SERVICE_KEY_PRV_KEY_PASS))
+		return str;
+
 	return g_strchomp(str);
 }
 
@@ -977,8 +1093,7 @@ static char *config_pem_fsid(const char *pem_file)
 	return g_strdup_printf("%llx", fsid64);
 }
 
-static void provision_service_wifi(gpointer key,
-				struct connman_config_service *config,
+static void provision_service_wifi(struct connman_config_service *config,
 				struct connman_service *service,
 				struct connman_network *network,
 				const void *ssid, unsigned int ssid_len)
@@ -990,10 +1105,30 @@ static void provision_service_wifi(gpointer key,
 		__connman_service_set_string(service, "Identity",
 							config->identity);
 
+	if (config->anonymous_identity)
+		__connman_service_set_string(service, "AnonymousIdentity",
+						config->anonymous_identity);
+
 	if (config->ca_cert_file)
 		__connman_service_set_string(service, "CACertFile",
 							config->ca_cert_file);
 
+	if (config->subject_match)
+		__connman_service_set_string(service, "SubjectMatch",
+							config->subject_match);
+
+	if (config->altsubject_match)
+		__connman_service_set_string(service, "AltSubjectMatch",
+							config->altsubject_match);
+
+	if (config->domain_suffix_match)
+		__connman_service_set_string(service, "DomainSuffixMatch",
+							config->domain_suffix_match);
+
+	if (config->domain_match)
+		__connman_service_set_string(service, "DomainMatch",
+							config->domain_match);
+
 	if (config->client_cert_file)
 		__connman_service_set_string(service, "ClientCertFile",
 						config->client_cert_file);
@@ -1055,42 +1190,71 @@ static gboolean remove_virtual_config(gpointer user_data)
 	return FALSE;
 }
 
-static void provision_service(gpointer key, gpointer value,
-							gpointer user_data)
+static int try_provision_service(struct connman_config_service *config,
+				struct connman_service *service)
 {
-	struct connman_service *service = user_data;
-	struct connman_config_service *config = value;
 	struct connman_network *network;
 	const void *service_id;
 	enum connman_service_type type;
 	const void *ssid;
 	unsigned int ssid_len;
-
-	type = connman_service_get_type(service);
-	if (type == CONNMAN_SERVICE_TYPE_WIFI &&
-				g_strcmp0(config->type, "wifi") != 0)
-		return;
-
-	if (type == CONNMAN_SERVICE_TYPE_ETHERNET &&
-				g_strcmp0(config->type, "ethernet") != 0)
-		return;
-
-	if (type == CONNMAN_SERVICE_TYPE_GADGET &&
-				g_strcmp0(config->type, "gadget") != 0)
-		return;
-
-	DBG("service %p ident %s", service,
-					__connman_service_get_ident(service));
+	const char *str;
 
 	network = __connman_service_get_network(service);
 	if (!network) {
 		connman_error("Service has no network set");
-		return;
+		return -EINVAL;
 	}
 
 	DBG("network %p ident %s", network,
 				connman_network_get_identifier(network));
 
+	type = connman_service_get_type(service);
+
+	switch(type) {
+	case CONNMAN_SERVICE_TYPE_WIFI:
+		if (__connman_service_string2type(config->type) != type)
+			return -ENOENT;
+
+		ssid = connman_network_get_blob(network, "WiFi.SSID",
+						&ssid_len);
+		if (!ssid)
+			return -ENOENT;
+
+		if (!config->ssid || ssid_len != config->ssid_len)
+			return -ENOENT;
+
+		if (memcmp(config->ssid, ssid, ssid_len))
+			return -ENOENT;
+
+		str = connman_network_get_string(network, "WiFi.Security");
+		if (config->security != __connman_service_string2security(str))
+			return -ENOENT;
+
+		break;
+
+	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_GADGET:
+
+		if (__connman_service_string2type(config->type) != type)
+			return -ENOENT;
+
+		break;
+
+	case CONNMAN_SERVICE_TYPE_UNKNOWN:
+	case CONNMAN_SERVICE_TYPE_SYSTEM:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_CELLULAR:
+	case CONNMAN_SERVICE_TYPE_GPS:
+	case CONNMAN_SERVICE_TYPE_VPN:
+	case CONNMAN_SERVICE_TYPE_P2P:
+
+		return -ENOENT;
+	}
+
+	DBG("service %p ident %s", service,
+				connman_service_get_identifier(service));
+
 	if (config->mac) {
 		struct connman_device *device;
 		const char *device_addr;
@@ -1098,7 +1262,7 @@ static void provision_service(gpointer key, gpointer value,
 		device = connman_network_get_device(network);
 		if (!device) {
 			connman_error("Network device is missing");
-			return;
+			return -ENODEV;
 		}
 
 		device_addr = connman_device_get_string(device, "Address");
@@ -1106,23 +1270,7 @@ static void provision_service(gpointer key, gpointer value,
 		DBG("wants %s has %s", config->mac, device_addr);
 
 		if (g_ascii_strcasecmp(device_addr, config->mac) != 0)
-			return;
-	}
-
-	if (g_strcmp0(config->type, "wifi") == 0 &&
-				type == CONNMAN_SERVICE_TYPE_WIFI) {
-		ssid = connman_network_get_blob(network, "WiFi.SSID",
-						&ssid_len);
-		if (!ssid) {
-			connman_error("Network SSID not set");
-			return;
-		}
-
-		if (!config->ssid || ssid_len != config->ssid_len)
-			return;
-
-		if (memcmp(config->ssid, ssid, ssid_len) != 0)
-			return;
+			return -ENOENT;
 	}
 
 	if (!config->ipv6_address) {
@@ -1140,12 +1288,12 @@ static void provision_service(gpointer key, gpointer value,
 
 		if (config->ipv6_prefix_length == 0) {
 			DBG("IPv6 prefix missing");
-			return;
+			return -EINVAL;
 		}
 
 		address = connman_ipaddress_alloc(AF_INET6);
 		if (!address)
-			return;
+			return -ENOENT;
 
 		connman_ipaddress_set_ipv6(address, config->ipv6_address,
 					config->ipv6_prefix_length,
@@ -1185,12 +1333,12 @@ static void provision_service(gpointer key, gpointer value,
 
 		if (!config->ipv4_netmask) {
 			DBG("IPv4 netmask missing");
-			return;
+			return -EINVAL;
 		}
 
 		address = connman_ipaddress_alloc(AF_INET);
 		if (!address)
-			return;
+			return -ENOENT;
 
 		connman_ipaddress_set_ipv4(address, config->ipv4_address,
 					config->ipv4_netmask,
@@ -1208,14 +1356,11 @@ static void provision_service(gpointer key, gpointer value,
 
 	__connman_service_disconnect(service);
 
-	service_id = __connman_service_get_ident(service);
+	service_id = connman_service_get_identifier(service);
 	config->service_identifiers =
 		g_slist_prepend(config->service_identifiers,
 				g_strdup(service_id));
 
-	if (!config->virtual)
-		__connman_service_set_immutable(service, true);
-
 	__connman_service_set_favorite_delayed(service, true, true);
 
 	__connman_service_set_config(service, config->config_ident,
@@ -1239,21 +1384,20 @@ static void provision_service(gpointer key, gpointer value,
 		__connman_service_set_search_domains(service,
 						config->search_domains);
 
+	__connman_service_set_mdns(service, config->mdns);
+
 	if (config->timeservers)
 		__connman_service_set_timeservers(service,
 						config->timeservers);
 
-	if (g_strcmp0(config->type, "wifi") == 0 &&
-				type == CONNMAN_SERVICE_TYPE_WIFI) {
-		provision_service_wifi(key, config, service, network,
+	if (type == CONNMAN_SERVICE_TYPE_WIFI) {
+		provision_service_wifi(config, service, network,
 							ssid, ssid_len);
-	} else
-		__connman_service_connect(service,
-					CONNMAN_SERVICE_CONNECT_REASON_AUTO);
+	}
 
 	__connman_service_mark_dirty();
 
-	__connman_service_save(service);
+	__connman_service_load_modifiable(service);
 
 	if (config->virtual) {
 		struct connect_virtual *virtual;
@@ -1262,16 +1406,63 @@ static void provision_service(gpointer key, gpointer value,
 		virtual->service = service;
 		virtual->vfile = config->virtual_file;
 
-		g_timeout_add(0, remove_virtual_config, virtual);
-	} else
-		__connman_service_auto_connect(CONNMAN_SERVICE_CONNECT_REASON_AUTO);
+		g_idle_add(remove_virtual_config, virtual);
+
+		return 0;
+	}
+
+	__connman_service_set_immutable(service, true);
+
+	if (type == CONNMAN_SERVICE_TYPE_ETHERNET ||
+			type == CONNMAN_SERVICE_TYPE_GADGET) {
+		__connman_service_connect(service,
+				CONNMAN_SERVICE_CONNECT_REASON_AUTO);
+
+		return 0;
+	}
+
+	__connman_service_auto_connect(CONNMAN_SERVICE_CONNECT_REASON_AUTO);
+
+	return 0;
+}
+
+static int
+find_and_provision_service_from_config(struct connman_service *service,
+					struct connman_config *config)
+{
+	GHashTableIter iter;
+	gpointer value, key;
+
+	g_hash_table_iter_init(&iter, config->service_table);
+	while (g_hash_table_iter_next(&iter, &key,
+					&value)) {
+		if (!try_provision_service(value, service))
+			return 0;
+	}
+
+	return -ENOENT;
+}
+
+static int find_and_provision_service(struct connman_service *service)
+{
+	GHashTableIter iter;
+	gpointer value, key;
+
+	g_hash_table_iter_init(&iter, config_table);
+
+	while (g_hash_table_iter_next(&iter, &key, &value)) {
+		struct connman_config *config = value;
+
+		if (!find_and_provision_service_from_config(service, config))
+			return 0;
+	}
+
+	return -ENOENT;
 }
 
 int __connman_config_provision_service(struct connman_service *service)
 {
 	enum connman_service_type type;
-	GHashTableIter iter;
-	gpointer value, key;
 
 	/* For now only WiFi, Gadget and Ethernet services are supported */
 	type = connman_service_get_type(service);
@@ -1283,16 +1474,7 @@ int __connman_config_provision_service(struct connman_service *service)
 			type != CONNMAN_SERVICE_TYPE_GADGET)
 		return -ENOSYS;
 
-	g_hash_table_iter_init(&iter, config_table);
-
-	while (g_hash_table_iter_next(&iter, &key, &value)) {
-		struct connman_config *config = value;
-
-		g_hash_table_foreach(config->service_table,
-						provision_service, service);
-	}
-
-	return 0;
+	return find_and_provision_service(service);
 }
 
 int __connman_config_provision_service_ident(struct connman_service *service,
@@ -1356,8 +1538,7 @@ int __connman_config_provision_service_ident(struct connman_service *service,
 			}
 		}
 
-		g_hash_table_foreach(config->service_table,
-						provision_service, service);
+		find_and_provision_service_from_config(service, config);
 	}
 
 	return ret;
@@ -1367,12 +1548,14 @@ static void generate_random_string(char *str, int length)
 {
 	uint8_t val;
 	int i;
+	uint64_t rand;
 
 	memset(str, '\0', length);
 
 	for (i = 0; i < length-1; i++) {
 		do {
-			val = (uint8_t)(random() % 122);
+			__connman_util_get_random(&rand);
+			val = (uint8_t)(rand % 122);
 			if (val < 48)
 				val += 48;
 		} while((val > 57 && val < 65) || (val > 90 && val < 97));
@@ -1385,7 +1568,7 @@ int connman_config_provision_mutable_service(GKeyFile *keyfile)
 {
 	struct connman_config_service *service_config;
 	struct connman_config *config;
-	char *vfile, *group;
+	char *vfile, *group = NULL;
 	char rstr[11];
 
 	DBG("");
@@ -1421,13 +1604,14 @@ int connman_config_provision_mutable_service(GKeyFile *keyfile)
 	if (g_strcmp0(service_config->type, "wifi") == 0)
 		__connman_device_request_scan(CONNMAN_SERVICE_TYPE_WIFI);
 
+	g_free(group);
 	return 0;
 
 error:
 	DBG("Could not proceed");
 	g_hash_table_remove(config_table, vfile);
 	g_free(vfile);
-
+	g_free(group);
 	return -EINVAL;
 }
 
@@ -1441,13 +1625,16 @@ struct connman_config_entry **connman_config_get_entries(const char *type)
 	g_hash_table_iter_init(&iter_file, config_table);
 	while (g_hash_table_iter_next(&iter_file, &key, &value)) {
 		struct connman_config *config_file = value;
+		struct connman_config_entry **tmp_entries = entries;
 
 		count = g_hash_table_size(config_file->service_table);
 
 		entries = g_try_realloc(entries, (i + count + 1) *
 					sizeof(struct connman_config_entry *));
-		if (!entries)
+		if (!entries) {
+			g_free(tmp_entries);
 			return NULL;
+		}
 
 		g_hash_table_iter_init(&iter_config,
 						config_file->service_table);
@@ -1480,10 +1667,14 @@ struct connman_config_entry **connman_config_get_entries(const char *type)
 	}
 
 	if (entries) {
+		struct connman_config_entry **tmp_entries = entries;
+
 		entries = g_try_realloc(entries, (i + 1) *
 					sizeof(struct connman_config_entry *));
-		if (!entries)
+		if (!entries) {
+			g_free(tmp_entries);
 			return NULL;
+		}
 
 		entries[i] = NULL;
 
@@ -1512,7 +1703,6 @@ void connman_config_free_entries(struct connman_config_entry **entries)
 	}
 
 	g_free(entries);
-	return;
 }
 
 bool __connman_config_address_provisioned(const char *address,
