@@ -59,10 +59,11 @@ struct connman_dhcp {
 	GDHCPClient *dhcp_client;
 	char *ipv4ll_debug_prefix;
 	char *dhcp_debug_prefix;
+
+	bool ipv4ll_running;
 };
 
 static GHashTable *ipconfig_table;
-static bool ipv4ll_running;
 
 static void dhcp_free(struct connman_dhcp *dhcp)
 {
@@ -85,7 +86,7 @@ static void ipv4ll_stop_client(struct connman_dhcp *dhcp)
 	g_dhcp_client_stop(dhcp->ipv4ll_client);
 	g_dhcp_client_unref(dhcp->ipv4ll_client);
 	dhcp->ipv4ll_client = NULL;
-	ipv4ll_running = false;
+	dhcp->ipv4ll_running = false;
 
 	g_free(dhcp->ipv4ll_debug_prefix);
 	dhcp->ipv4ll_debug_prefix = NULL;
@@ -232,7 +233,7 @@ static int ipv4ll_start_client(struct connman_dhcp *dhcp)
 		return err;
 	}
 
-	ipv4ll_running = true;
+	dhcp->ipv4ll_running = true;
 	return 0;
 }
 
@@ -253,7 +254,7 @@ static void no_lease_cb(GDHCPClient *dhcp_client, gpointer user_data)
 	struct connman_dhcp *dhcp = user_data;
 	int err;
 
-	DBG("No lease available ipv4ll %d client %p", ipv4ll_running,
+	DBG("No lease available ipv4ll %d client %p", dhcp->ipv4ll_running,
 		dhcp->ipv4ll_client);
 
 	if (dhcp->timeout > 0)
@@ -262,7 +263,7 @@ static void no_lease_cb(GDHCPClient *dhcp_client, gpointer user_data)
 	dhcp->timeout = g_timeout_add_seconds(RATE_LIMIT_INTERVAL,
 						dhcp_retry_cb,
 						dhcp);
-	if (ipv4ll_running)
+	if (dhcp->ipv4ll_running)
 		return;
 
 	err = ipv4ll_start_client(dhcp);
@@ -270,7 +271,7 @@ static void no_lease_cb(GDHCPClient *dhcp_client, gpointer user_data)
 		DBG("Cannot start ipv4ll client (%d/%s)", err, strerror(-err));
 
 	/* Only notify upper layer if we have a problem */
-	dhcp_invalidate(dhcp, !ipv4ll_running);
+	dhcp_invalidate(dhcp, !dhcp->ipv4ll_running);
 }
 
 static void lease_lost_cb(GDHCPClient *dhcp_client, gpointer user_data)
@@ -435,6 +436,7 @@ static void lease_available_cb(GDHCPClient *dhcp_client, gpointer user_data)
 {
 	struct connman_dhcp *dhcp = user_data;
 	GList *option = NULL;
+	enum connman_ipconfig_method old_method;
 	char *address, *netmask = NULL, *gateway = NULL;
 	const char *c_address, *c_gateway;
 	unsigned char prefixlen, c_prefixlen;
@@ -486,8 +488,24 @@ static void lease_available_cb(GDHCPClient *dhcp_client, gpointer user_data)
 	} else if (prefixlen != c_prefixlen)
 		ip_change = true;
 
+	old_method = __connman_ipconfig_get_method(dhcp->ipconfig);
 	__connman_ipconfig_set_method(dhcp->ipconfig,
 						CONNMAN_IPCONFIG_METHOD_DHCP);
+
+	/*
+	 * Notify IPv4.Configuration's method moved back to DHCP.
+	 *
+	 * This is the case ConnMan initially set an address by using
+	 * IPv4LL because DHCP failed but now we got an address from DHCP.
+	 */
+	if (old_method == CONNMAN_IPCONFIG_METHOD_AUTO) {
+		struct connman_service *service =
+			connman_service_lookup_from_network(dhcp->network);
+
+		if (service)
+			__connman_service_notify_ipv4_configuration(service);
+	}
+
 	if (ip_change) {
 		__connman_ipconfig_set_local(dhcp->ipconfig, address);
 		__connman_ipconfig_set_prefixlen(dhcp->ipconfig, prefixlen);
@@ -509,6 +527,7 @@ done:
 static void ipv4ll_available_cb(GDHCPClient *ipv4ll_client, gpointer user_data)
 {
 	struct connman_dhcp *dhcp = user_data;
+	enum connman_ipconfig_method old_method;
 	char *address, *netmask;
 	unsigned char prefixlen;
 
@@ -519,8 +538,25 @@ static void ipv4ll_available_cb(GDHCPClient *ipv4ll_client, gpointer user_data)
 
 	prefixlen = connman_ipaddress_calc_netmask_len(netmask);
 
+	old_method = __connman_ipconfig_get_method(dhcp->ipconfig);
 	__connman_ipconfig_set_method(dhcp->ipconfig,
-						CONNMAN_IPCONFIG_METHOD_DHCP);
+						CONNMAN_IPCONFIG_METHOD_AUTO);
+
+	/*
+	 * Notify IPv4.Configuration's method is AUTO now.
+	 *
+	 * This is the case DHCP failed thus ConnMan used IPv4LL to get an
+	 * address. Set IPv4.Configuration method to AUTO allows user to
+	 * ask for a DHCP address by setting the method again to DHCP.
+	 */
+	if (old_method == CONNMAN_IPCONFIG_METHOD_DHCP) {
+		struct connman_service *service =
+			connman_service_lookup_from_network(dhcp->network);
+
+		if (service)
+			__connman_service_notify_ipv4_configuration(service);
+	}
+
 	__connman_ipconfig_set_local(dhcp->ipconfig, address);
 	__connman_ipconfig_set_prefixlen(dhcp->ipconfig, prefixlen);
 	__connman_ipconfig_set_gateway(dhcp->ipconfig, NULL);
@@ -709,6 +745,30 @@ void __connman_dhcp_stop(struct connman_ipconfig *ipconfig)
 	}
 }
 
+void __connman_dhcp_decline(struct connman_ipconfig *ipconfig)
+{
+	struct connman_dhcp *dhcp;
+	const char *address;
+	struct in_addr addr;
+
+	DBG("ipconfig_table %p ipconfig %p", ipconfig_table, ipconfig);
+
+	if (!ipconfig_table)
+		return;
+
+	dhcp = g_hash_table_lookup(ipconfig_table, ipconfig);
+	if (dhcp) {
+		address = __connman_ipconfig_get_local(ipconfig);
+		if (!address)
+			return;
+
+		if (inet_pton(AF_INET, address, &addr) != 1)
+			connman_error("Could not convert address %s", address);
+
+		g_dhcp_client_decline(dhcp->dhcp_client, htonl(addr.s_addr));
+	}
+}
+
 int __connman_dhcp_init(void)
 {
 	DBG("");
@@ -725,6 +785,4 @@ void __connman_dhcp_cleanup(void)
 
 	g_hash_table_destroy(ipconfig_table);
 	ipconfig_table = NULL;
-
-	dhcp_cleanup_random();
 }
